from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import os
import binascii
import hashlib
import oz.app

from .middleware import *
from .options import *
from .tests import *

def random_hex(length):
    """Generates a random hex string"""
    return binascii.hexlify(os.urandom(length))[length:]

def password_hash(password, password_salt=None):
    """Hashes a specified password"""
    password_salt = password_salt or oz.app.settings["session_salt"]
    return u"sha256!%s" % hashlib.sha256(unicode(password_salt) + unicode(password)).hexdigest()



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz.plugins.bandit
import oz.plugins.redis
import unittest

@oz.test
class ChiSquaredTest(unittest.TestCase):
    def test_get_chi_squared(self):

        # Make sure it returns 0 for an experiment with no results
        csq = oz.plugins.bandit.get_chi_squared(0, 0, 0, 0)
        self.assertEqual(csq, 0)

        csq = oz.plugins.bandit.get_chi_squared(17, 3, 19, 2)
        self.assertEqual(round(csq, 3), 0.328)

@oz.test
class BanditCoreTest(unittest.TestCase):
    def tearDown(self):
        super(BanditCoreTest, self).tearDown()
        
        # Clean up just in case there's keys still lying around
        redis = oz.plugins.redis.create_connection()

        for key in redis.keys("bandit:*:v2"):
            redis.delete(key)

    def test_exists(self):
        redis = oz.plugins.redis.create_connection()
        experiment = oz.plugins.bandit.Experiment(redis, "ex-exists")
        self.assertFalse(experiment.exists())
        experiment.add()
        self.assertTrue(experiment.exists())

    def test_archive(self):
        redis = oz.plugins.redis.create_connection()
        experiment = oz.plugins.bandit.Experiment(redis, "ex-archived")

        self.assertEqual(redis.scard(oz.plugins.bandit.ACTIVE_EXPERIMENTS_REDIS_KEY), 0)
        self.assertEqual(redis.scard(oz.plugins.bandit.ARCHIVED_EXPERIMENTS_REDIS_KEY), 0)

        experiment.add()
        self.assertEqual(redis.scard(oz.plugins.bandit.ACTIVE_EXPERIMENTS_REDIS_KEY), 1)
        self.assertEqual(redis.scard(oz.plugins.bandit.ARCHIVED_EXPERIMENTS_REDIS_KEY), 0)

        experiment.archive()
        self.assertEqual(redis.scard(oz.plugins.bandit.ACTIVE_EXPERIMENTS_REDIS_KEY), 0)
        self.assertEqual(redis.scard(oz.plugins.bandit.ARCHIVED_EXPERIMENTS_REDIS_KEY), 1)

    def test_metadata(self):
        redis = oz.plugins.redis.create_connection()
        experiment = oz.plugins.bandit.Experiment(redis, "ex-metadata")
        self.assertEqual(experiment.metadata(), None)
        experiment.add()
        self.assertNotEqual(experiment.metadata(), None)

    def test_choices(self):
        redis = oz.plugins.redis.create_connection()
        experiment = oz.plugins.bandit.Experiment(redis, "ex-choices")

        self.assertRaises(lambda: experiment.add_choice("illegal name"))
        self.assertEqual(experiment.choices(), [])

        experiment.add_choice("A")
        experiment.add_choice("B")
        self.assertEqual(experiment.choices(), ["A", "B"])

        experiment.remove_choice("B")
        self.assertEqual(experiment.choices(), ["A"])

    def test_plays(self):
        redis = oz.plugins.redis.create_connection()

        experiment = oz.plugins.bandit.Experiment(redis, "ex-plays")
        experiment.add()

        # Make sure there are no plays
        self.assertEqual(experiment.plays("foo"), 0)

        # Add a play
        experiment.add_play("foo")

        # Make sure the play exists
        self.assertEqual(experiment.plays("foo"), 1)

    def test_rewards(self):
        redis = oz.plugins.redis.create_connection()

        experiment = oz.plugins.bandit.Experiment(redis, "ex-rewards")
        experiment.add()

        # Make sure there are no rewards
        self.assertEqual(experiment.rewards("foo"), 0)

        # Add a reward
        experiment.add_reward("foo")

        # Make sure the reward exists
        self.assertEqual(experiment.rewards("foo"), 1)

    def test_default_choice(self):
        redis = oz.plugins.redis.create_connection()

        experiment = oz.plugins.bandit.Experiment(redis, "ex-default-choice")
        experiment.add()

        self.assertEqual(experiment.get_default_choice(), None)
        experiment.set_default_choice("asd")
        self.assertEqual(experiment.get_default_choice(), "asd")

    def test_results(self):
        redis = oz.plugins.redis.create_connection()

        experiment = oz.plugins.bandit.Experiment(redis, "ex-data")
        experiment.add()

        experiment.add_choice("arm1")
        experiment.add_choice("arm2")

        # Add data to the experiment
        experiment.add_play("arm1", count=17)
        experiment.add_reward("arm1", count=3)
        experiment.add_play("arm2", count=19)
        experiment.add_reward("arm2", count=2)

        data = experiment.results()

        self.assertEqual(data["name"], "ex-data")
        self.assertNotEqual(data["metadata"], None)
        self.assertEqual(data["default"], "arm1")
        self.assertEqual(round(data["chi_squared"], 3), 0.328)

        self.assertEqual(len(data["choices"]), 2)

        arm1 = data["choices"][0]
        self.assertEqual(arm1["name"], "arm1")
        self.assertEqual(round(arm1["results"], 3), 0.176)
        self.assertEqual(arm1["plays"], 17)
        self.assertEqual(arm1["rewards"], 3)

        arm2 = data["choices"][1]
        self.assertEqual(arm2["name"], "arm2")
        self.assertEqual(round(arm2["results"], 3), 0.105)
        self.assertEqual(arm2["plays"], 19)
        self.assertEqual(arm2["rewards"], 2)



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import datetime

import oz.plugins.redis
import oz.plugins.bandit

@oz.action
def add_experiment(experiment):
    """Adds a new experiment"""
    redis = oz.plugins.redis.create_connection()
    oz.plugins.bandit.Experiment(redis, experiment).add()

@oz.action
def archive_experiment(experiment):
    """Archives an experiment"""
    redis = oz.plugins.redis.create_connection()
    oz.plugins.bandit.Experiment(redis, experiment).archive()

@oz.action
def add_experiment_choice(experiment, choice):
    """Adds an experiment choice"""
    redis = oz.plugins.redis.create_connection()
    oz.plugins.bandit.Experiment(redis, experiment).add_choice(choice)

@oz.action
def remove_experiment_choice(experiment, choice):
    """Removes an experiment choice"""
    redis = oz.plugins.redis.create_connection()
    oz.plugins.bandit.Experiment(redis, experiment).remove_choice(choice)

@oz.action
def get_experiment_results():
    """
    Computes the results of all experiments, stores it in redis, and prints it
    out
    """

    redis = oz.plugins.redis.create_connection()

    for experiment in oz.plugins.bandit.get_experiments(redis):
        data = experiment.results()

        print("%s:" % data["name"])
        print("- creation date: %s" % data["metadata"]["creation_date"])
        print("- default choice: %s" % data["default"])
        print("- chi squared: %s" % data["chi_squared"])
        print("- choices:")

        for choice_data in data["choices"]:
            print("  - %s: plays=%s, rewards=%s, result=%s" % (choice_data["name"], choice_data["plays"], choice_data["rewards"], choice_data["results"]))



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import redis
import oz.app

from .middleware import *
from .options import *

_cached_connection = None

def create_connection():
    """Sets up a redis configuration"""

    global _cached_connection
    settings = oz.app.settings

    if settings["redis_cache_connections"] and _cached_connection != None:
        return _cached_connection
    else:
        conn = redis.StrictRedis(
            host=settings["redis_host"],
            port=settings["redis_port"],
            db=settings["redis_db"],
            password=settings["redis_password"]
        )

        if settings["redis_cache_connections"]:
            _cached_connection = conn

        return conn



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.app
import re
import oz.plugins.json_api
from tornado import escape
import traceback

API_ERROR_CODE_MAP = { 404: "Not found", 500: "Server error" }
CALLBACK_VALIDATOR = re.compile('^[$A-Za-z_][0-9A-Za-z_$]*$')

class ApiMiddleware(object):
    def __init__(self):
        super(ApiMiddleware, self).__init__()
        self._decoded_body = None
        self.trigger_listener("write_error", self._api_on_write_error)

    def _api_on_write_error(self, status_code, **kwargs):
        """
        Catches errors and renders it as a JSON message. Adds the traceback if
        debug is enabled.
        """

        return_error = { "code": self.get_status() }
        exc_info = kwargs.get("exc_info")

        if exc_info and isinstance(exc_info[1], oz.plugins.json_api.ApiError):
            return_error["error"] = exc_info[1].message
        else:
            return_error["error"] = API_ERROR_CODE_MAP.get(self.get_status(), "Unknown error")

        if oz.app.settings.get("debug"):
            return_error["trace"] = "".join(traceback.format_exception(*exc_info))

        self.finish(return_error)
        return oz.break_trigger

    def respond(self, obj):
        """Gives a response JSON(P) message"""

        # Get the callback argument if JSONP is allowed
        callback = self.get_argument("callback", None) if oz.app.settings["allow_jsonp"] else None

        # We're pretty strict with what callback names are allowed, just in case
        if callback and not CALLBACK_VALIDATOR.match(callback):
            raise oz.plugins.json_api.ApiError("Invalid callback identifier - only functions with ASCII characters are allowed")

        # Provide the response in a different manner depending on whether a
        # JSONP callback is specified
        json = escape.json_encode(obj)

        if callback:
            self.set_header("Content-Type", "application/javascript; charset=UTF-8")
            self.finish("%s(%s)" % (callback, json))
        else:
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            self.finish(json)

    def body(self):
        """Gets the JSON body of the request"""

        if self._decoded_body == None:
            # Try to decode the JSON body. But raise an error if the
            # content-type is unexpected, or the JSON is invalid.

            raw_content_type = self.request.headers.get("content-type") or ""
            content_type = raw_content_type.split(";")[0].strip().lower()

            if content_type == "application/json":
                try:
                    self._decoded_body = escape.json_decode(self.request.body)
                except:
                    raise oz.plugins.json_api.ApiError("Bad JSON body")
            else:
                raise oz.plugins.json_api.ApiError("JSON body expected")

        return self._decoded_body



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import tornado.web
import tornado.wsgi
import tornado.ioloop
import tornado.httpserver
import wsgiref.simple_server
import oz
import oz.app
import oz.testing
import sys
import shutil
import unittest

@oz.action
def server():
    """Runs the server"""

    settings = oz.app.settings

    if settings["wsgi_mode"]:
        application = tornado.wsgi.WSGIApplication(oz.app.routes, **settings)
        server = wsgiref.simple_server.make_server("", settings["port"], application)
        server.serve_forever()
    else:
        application = tornado.web.Application(oz.app.routes, **settings)
        server = tornado.httpserver.HTTPServer(application)
        server.bind(settings["port"])

        if settings["debug"]:
            if settings["server_workers"] != 1:
                print >> sys.stderr, "WARNING: Debug is enabled, but multiple server workers have been configured. Only one server worker can run in debug mode."

            server.start(1)
        else:
            # Forks multiple sub-processes
            server.start(settings["server_workers"])
        
        tornado.ioloop.IOLoop.instance().start()

@oz.action
def repl():
    """Runs an IPython repl with some context"""

    try:
        import IPython
    except:
        print("ERROR: IPython is not installed. Please install it to use the repl.", file=sys.stderr)
        raise

    IPython.embed(user_ns=dict(
        settings=oz.app.settings,
        actions=oz.app.actions,
        uimodules=oz.app.uimodules,
        routes=oz.app.routes,
    ))

@oz.action
def test(*filters):
    """Runs the unit tests"""

    suite = unittest.TestSuite()
    filters_set = set(filters)

    for test in oz.app.tests:
        if not filters_set or test.__name__ in filters_set:
            # Add the class to the suite
            child_suite = unittest.makeSuite(test, "test")
            suite.addTest(child_suite)

    res = unittest.TextTestRunner().run(suite)
    return 1 if len(res.errors) > 0 or len(res.failures) > 0 else 0



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.testing
import oz.plugins.blinks

@oz.test
class BlinkMiddlewareTest(oz.testing.OzTestCase):
    def get_handlers(self):
        class BlinkHandler(oz.testing.FakeCookiesHandler, oz.plugins.blinks.BlinkMiddleware):
            def get(self):
                self.set_blink("Blink 1")
                self.set_blink("Blink #?", type="new_blink")
                self.finish("%s : %s" % self.get_blink())

        return [
            ("/", BlinkHandler),
        ]

    def test_blink(self):
        self.http_client.fetch(self.get_url('/'), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "Blink #? : new_blink")



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.testing
import collections
from oz.plugins.redis_sessions import RedisSessionMiddleware
from oz.plugins.redis import RedisMiddleware

@oz.test
class RedisSessionMiddlewareTest(oz.testing.OzTestCase):
    forced_settings = {
        "session_salt": "abc",
        "session_time": 60 * 1000
    }

    def get_handlers(self):
        class SessionGetterHandler(oz.testing.FakeCookiesHandler, RedisMiddleware, RedisSessionMiddleware):
            def get(self):
                self.finish(self.get_session_value(self.get_argument("name"), self.get_argument("default", None)))

        class SessionSetterHandler(oz.testing.FakeCookiesHandler, RedisMiddleware, RedisSessionMiddleware):
            def get(self):
                self.finish(self.set_session_value(self.get_argument("name"), self.get_argument("value")))

        class SessionDeleterHandler(oz.testing.FakeCookiesHandler, RedisMiddleware, RedisSessionMiddleware):
            def get(self):
                self.finish(self.clear_session_value(self.get_argument("name")))

        class SessionClearerHandler(oz.testing.FakeCookiesHandler, RedisMiddleware, RedisSessionMiddleware):
            def get(self):
                self.clear_all_session_values()

        return [
            ("/get", SessionGetterHandler),
            ("/set", SessionSetterHandler),
            ("/del", SessionDeleterHandler),
            ("/clear", SessionClearerHandler),
        ]

    def tearDown(self):
        super(RedisSessionMiddlewareTest, self).tearDown()
        
        # Clean up just in case there's keys still lying around
        redis = oz.plugins.redis.create_connection()

        for key in redis.keys("session:*:v1"):
            redis.delete(key)

    def test_session_value(self):
        self.http_client.fetch(self.get_url("/set?name=session_value_1&value=bar&cookie_id=test_session_value"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")

        self.http_client.fetch(self.get_url("/get?name=session_value_1&cookie_id=test_session_value"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "bar")

        self.http_client.fetch(self.get_url("/get?name=session_value_2&cookie_id=test_session_value"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")

        self.http_client.fetch(self.get_url("/get?name=session_value_2&default=bar2&cookie_id=test_session_value"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "bar2")

    def test_clear_session_value(self):
        self.http_client.fetch(self.get_url("/set?name=clearable_value&value=baz&cookie_id=test_clear_session_value"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")

        self.http_client.fetch(self.get_url("/del?name=clearable_value&cookie_id=test_clear_session_value"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")

        self.http_client.fetch(self.get_url("/get?name=clearable_value&cookie_id=test_clear_session_value"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")

    def test_clear_all_session_values(self):
        self.http_client.fetch(self.get_url("/set?name=clear_all_value_1&value=baz&cookie_id=test_clear_all_session_values"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")

        self.http_client.fetch(self.get_url("/set?name=clear_all_value_2&value=baz&cookie_id=test_clear_all_session_values"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")

        self.http_client.fetch(self.get_url("/clear?cookie_id=test_clear_all_session_values"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")

        self.http_client.fetch(self.get_url("/get?name=clear_all_value_1&cookie_id=test_clear_all_session_values"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")

        self.http_client.fetch(self.get_url("/get?name=clear_all_value_2&cookie_id=test_clear_all_session_values"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "")



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.testing
import collections
import oz.plugins.redis
from oz.plugins.aws_cdn import CDNMiddleware
from oz.plugins.redis import RedisMiddleware

@oz.test
class CDNMiddlewareTest(oz.testing.OzTestCase):
    forced_settings = {
        "static_host": "//fakecdn"
    }

    def get_handlers(self):
        class CDNHandler(oz.RequestHandler, RedisMiddleware, CDNMiddleware):
            def put(self):
                file, value = self.request.body.split(":")
                self.set_cache_buster(file, value)
                self.finish("ok")

        class CacheBusterHandler(CDNHandler):
            def get(self):
                self.finish(self.get_cache_buster(self.get_argument("file")))

        class StaticUrlHandler(CDNHandler):
            def get(self):
                self.finish(self.cdn_static_url(self.get_argument("file")))

        return [
            ("/cache_buster", CacheBusterHandler),
            ("/static_url", StaticUrlHandler),
        ]

    def tearDown(self):
        super(CDNMiddlewareTest, self).tearDown()

        # Kill the created key
        redis = oz.plugins.redis.create_connection()
        redis.delete("cache-buster:v1")

    def test_get_cache_buster(self):
        self.http_client.fetch(self.get_url("/cache_buster"), self.stop, method="PUT", body="path/to/file.txt:abcdef")
        response = self.wait()
        self.assertEqual(response.body, "ok")

        self.http_client.fetch(self.get_url("/cache_buster?file=path%2Fto%2Ffile.txt"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "abcdef")

    def test_static_url(self):
        self.http_client.fetch(self.get_url("/static_url"), self.stop, method="PUT", body="path/to/file2.txt:ghijkl")
        response = self.wait()
        self.assertEqual(response.body, "ok")

        self.http_client.fetch(self.get_url("/static_url?file=path%2Fto%2Ffile2.txt"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "//fakecdn/path/to/file2.txt?v=ghijkl")



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import types
import collections
import tornado.web
import tornado.options
import inspect
import functools
import collections
from . import app

# On trigger execution, trigger listeners can return this to notify the
# request handler to cancel execution of the next functions in the trigger
# chain.
break_trigger = object()

def _add_to_dict(type, container, name, value):
    """
    Adds an item to a dictionary, or raises an exception if an item with the
    specified key already exists in the dictionary.
    """

    if name in container:
        raise Exception("%s '%s' already exists" % (type, name))
    else:
        container[name] = value

def action(fun):
    """Exposes an action"""
    _add_to_dict("Action", app.actions, fun.__name__, fun)
    return fun

def uimodule(cls):
    """Exposes a UIModule"""
    _add_to_dict("UIModules", app.uimodules, cls.__name__, cls)
    return cls

def route(new_route):
    """Exposes a route"""
    app.routes.append(new_route)

def routes(*new_routes):
    """Exposes a list of routes"""
    app.routes.extend(new_routes)

def option(name, **args):
    """Exposes an option"""
    _add_to_dict("Option", app.options, name, args)

def options(**kwargs):
    """Exposes several options"""
    for name, args in kwargs.items():
        option(name, **args)

def test(cls):
    """Exposes a unit test class, to be run on the `test` action."""
    app.tests.append(cls)
    return cls

def signal(name):
    """Adds a function to be executed on a signal"""
    def wrapper(fn):
        if not name in app.signals:
            app.signals[name] = []

        app.signals[name].append(fn)
        return fn

    return wrapper

class RequestHandler(tornado.web.RequestHandler):
    def __init__(self, *args, **kwargs):
        self._template_helpers = {}
        self._triggers = collections.defaultdict(lambda: [])
        super(RequestHandler, self).__init__(*args, **kwargs)

    def template_helper(self, name, callback):
        """Adds a template helper"""
        _add_to_dict("Template helper", self._template_helpers, name, callback)

    def trigger_listener(self, name, callback):
        """Adds a trigger listener"""
        self._triggers[name].append(callback)

    def trigger(self, name, *args, **kwargs):
        """
        Triggers an event to run through middleware. This method will execute
        a chain of relevant trigger callbacks, until one of the callbacks
        returns the `break_trigger`.
        """

        # Relevant middleware is cached so we don't have to rediscover it
        # every time. Fetch the cached value if possible.

        listeners = self._triggers.get(name, [])

        # Execute each piece of middleware
        for listener in listeners:
            result = listener(*args, **kwargs)

            if result == break_trigger:
                return False

        return True

    def get_template_namespace(self):
        # Start with the default namespace
        namespace = super(RequestHandler, self).get_template_namespace()

        # Add the template helpers
        namespace.update(self._template_helpers)

        return namespace

    def initialize(self, **kwargs):
        self.trigger("initialize", **kwargs)

    def prepare(self):
        self.trigger("prepare")

    def on_finish(self):
        self.trigger("on_finish")

    def on_connection_close(self):
        self.trigger("on_connection_close")

    def write_error(self, *args, **kwargs):
        # Call super's implementation of `write_error` unless we hit a
        # `break_trigger`.
        if self.trigger("write_error", *args, **kwargs):
            super(RequestHandler, self).write_error(*args, **kwargs)

def initialize(config):
    # Load the plugins
    for plugin in config.plugins:
        __import__(plugin, globals(), locals(), [], -1)
        execute_signal("plugin_loaded", plugin)

    # Add the options
    for option_name, option_kwargs in app.options.items():
        tornado.options.define(option_name, **option_kwargs)

    for key, value in config.app_options.items():
        setattr(tornado.options.options, key, value)

    # Generate the application settings
    settings = dict((key, getattr(tornado.options.options, key)) for key in app.options.keys())
    settings["project_name"] = config.project_name
    settings["ui_modules"] = app.uimodules
    app.settings = settings

    execute_signal("initialized")

def execute_signal(name, *args, **kwargs):
    """
    Executes a signal, running all of the associated callbacks with the given
    args/kwargs
    """

    for callback in app.signals.get(name, []):
        callback(*args, **kwargs)



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
from tornado import escape

class BlinkMiddleware(object):
    def __init__(self):
        super(BlinkMiddleware, self).__init__()
        self.template_helper("get_blink", self.get_blink)

    def get_blink_cookie(self, name):
        """Gets a blink cookie value"""
        value = self.get_cookie(name)

        if value != None:
            self.clear_cookie(name)
            return escape.url_unescape(value)

    def set_blink(self, message, type="info"):
        """
        Sets the blink, a one-time transactional message that is shown on the
        next page load
        """
        self.set_cookie("blink_message", escape.url_escape(message), httponly=True)
        self.set_cookie("blink_type", escape.url_escape(type), httponly=True)

    def get_blink(self):
        """Gets the blink message and type"""
        return self.get_blink_cookie("blink_message"), self.get_blink_cookie("blink_type")



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.plugins.sqlalchemy

@oz.action
def setup_database():
    """Sets up the database tables"""
    oz.plugins.sqlalchemy.setup()



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import os
import functools
import shutil
import re
import oz
import sys

VALID_PROJECT_NAME = re.compile("^\w+$")

def check_path(path, otherwise):
    """
    Checks if a path exists. If it does, print a warning message; if not,
    execute the `otherwise` callback argument.
    """

    if os.path.exists(path):
        print("WARNING: Path '%s' already exists; skipping" % path)
    else:
        otherwise(path)

def config_maker(project_name, path):
    """Creates a config file based on the project name"""

    with open(skeleton_path("config.py"), "r") as config_source:
        config_content = config_source.read()

    config_content = config_content.replace("__PROJECT_NAME__", project_name)

    with open(path, "w") as config_dest:
        config_dest.write(config_content)

def skeleton_path(parts):
    """Gets the path to a skeleton asset"""
    return os.path.join(os.path.dirname(oz.__file__), "skeleton", parts)

def init(project_name):
    """Creates a new project"""

    if not VALID_PROJECT_NAME.match(project_name):
        print("Invalid project name. It may only contain letters, numbers and underscores.", file=sys.stderr)
        return

    check_path(project_name, functools.partial(shutil.copytree, skeleton_path("plugin")))
    check_path("static", os.mkdir)
    check_path("templates", os.mkdir)
    check_path("config.py", functools.partial(config_maker, project_name))



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz

oz.options(
    debug = dict(type=bool, default=False, help="Debug mode. Stack traces will be displayed on errors, and the server will automatically reload when server code is changed. This should not be enabled in production."),
    port = dict(type=int, default=8000, help="The port on which to run the server."),
    static_path = dict(type=str, default=None, help="Path to the static directory. If set to None, then static asset hosting is disabled."),
    template_path = dict(type=str, default="templates", help="Path to the templates directory."),
    xsrf_cookies = dict(type=bool, default=False, help="Enable cookies that prevent XSRF attacks."),
    cookie_secret = dict(type=str, help="Secret used for secure cookies. Required if you set xsrf_cookies=True."),
    gzip = dict(type=bool, help="Enables gzip output."),
    wsgi_mode = dict(type=bool, default=False, help="Use WSGI mode."),
    server_workers = dict(type=int, default=1, help="The number of server workers to run concurrently.")
)


from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz

oz.options(
    session_salt = dict(type=str, help="Salt used for session security"),
    session_time = dict(type=int, help="Number of seconds of session inactivity before timeout")
)


from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import sys
import warnings
import oz.plugins.aws_cdn
import oz.app

class CDNMiddleware(object):
    def __init__(self):
        super(CDNMiddleware, self).__init__()
        self.template_helper("cdn_static_url", self.cdn_static_url)

    def cdn_static_url(self, path):
        """
        Takes a path and returns a static URL based on the CDN configuration.
        Largely a drop-in replacement for tornado's `static_url`.
        """
        return oz.plugins.aws_cdn.static_url(self.redis(), path)

    def get_cache_buster(self, path):
        """Gets the cache buster value for a given file path"""
        return oz.plugins.aws_cdn.get_cache_buster(self.redis(), path)

    def set_cache_buster(self, path, hash):
        """Sets the cache buster value for a given file path"""
        oz.plugins.aws_cdn.set_cache_buster(self.redis(), path, hash)

    def remove_cache_buster(self, path):
        """Removes the cache buster for a given file"""
        oz.plugins.aws_cdn.remove_cache_buster(self.redis(), path)

    def get_file(self, path):
        """Gets a file at the given path"""
        return oz.plugins.aws_cdn.get_file(path)

    def upload_file(self, path, contents, replace=False):
        """
        Uplodas the file to its path with the given `content`, adding the
        appropriate parent directories when needed. If the path already exists
        and `replace` is `False`, the file will not be uploaded.
        """
        f = self.get_file(path)
        f.upload(contents, replace=replace)
        self.set_cache_buster(path, f.hash(override=oz.app.settings.get("hash_override", "")))

    def copy_file(self, from_path, to_path, replace=False):
        """
        Copies a file from a given source path to a destination path, adding
        appropriate parent directories when needed. If the destination path
        already exists and `replace` is `False`, the file will not be
        uploaded.
        """
        f = self.get_file(from_path)
        self.upload_file(to_path, f.contents(), replace=replace)

    def remove_file(self, path):
        """Removes the given file"""
        f = self.get_file(path).remove()
        self.remove_cache_buster(path)



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

from tornado import escape
from .actions import *
from .middleware import *
from .tests import *
import re

ACTIVE_EXPERIMENTS_REDIS_KEY = "bandit:listing:active:v2"
ARCHIVED_EXPERIMENTS_REDIS_KEY = "bandit:listing:archived:v2"
EXPERIMENT_REDIS_KEY_TEMPLATE = "bandit:experiment:%s:v2"
ALLOWED_NAMES = re.compile("^[A-Za-z0-9-_]+$")

def get_chi_squared(choice1_plays, choice1_rewards, choice2_plays, choice2_rewards):
    """Calculates chi-squared between two choices"""

    # Currently not set up to calculate chi_squared for more than 2 arms
    exp_0 = (float(choice1_rewards) + choice2_rewards) / ((choice1_plays + choice2_plays) or 1) * choice1_plays
    exp_1 = (float(choice1_rewards) + choice2_rewards) / ((choice1_plays + choice2_plays) or 1) * choice2_plays

    if exp_0 > 0 and exp_1 > 0:
        return (exp_0 - float(choice1_rewards)) ** 2 / exp_0 + (exp_1 - float(choice2_rewards)) ** 2 / exp_1
    else:
        return 0

class ExperimentException(Exception):
    """
    An exception for issues related to experiments. This way applications can
    catch the exception and respond accordingly.
    """

    def __init__(self, experiment_name, message):
        super(ExperimentException, self).__init__("Experiment %s: %s" % (experiment_name, message))
        self.experiment_name = experiment_name

class Experiment(object):
    def __init__(self, redis, name):
        if not ALLOWED_NAMES.match(name):
            raise ExperimentException(name, "Illegal name")

        self.redis = redis
        self.name = name

    def _redis_key(self):
        return EXPERIMENT_REDIS_KEY_TEMPLATE % self.name

    def exists(self):
        """Checks whether this experiment exists"""
        return self.redis.exists(self._redis_key())

    def add(self):
        """Adds the new experiment"""

        # Check to ensure the experiment doesn't already exists
        if self.exists():
            raise ExperimentException(self.name, "already exists")

        # Add the experiment
        json = dict(creation_date=unicode(datetime.datetime.now()))
        pipe = self.redis.pipeline(transaction=True)
        pipe.sadd(ACTIVE_EXPERIMENTS_REDIS_KEY, self.name)
        pipe.hset(self._redis_key(), "metadata", escape.json_encode(json))
        pipe.execute()

    def archive(self):
        """Archives an experiment"""

        if not self.exists():
            raise ExperimentException(self.name, "does not exist") 

        pipe = self.redis.pipeline(transaction=True)
        pipe.srem(ACTIVE_EXPERIMENTS_REDIS_KEY, self.name)
        pipe.sadd(ARCHIVED_EXPERIMENTS_REDIS_KEY, self.name)
        pipe.execute()

    def metadata(self):
        """Gets the properties associated with this experiment"""
        raw = self.redis.hget(self._redis_key(), "metadata")
        return escape.json_decode(raw) if raw != None else None

    def choices(self):
        """Gets the choices available"""
        choices = self.redis.hget(self._redis_key(), "choices")
        return escape.json_decode(choices) if choices != None else []

    def add_choice(self, choice):
        """Adds a choice for the experiment"""

        if not ALLOWED_NAMES.match(choice):
            raise ExperimentException(self.name, "Illegal choice name: %s" % choice)

        choices = self.choices()
        choices.append(choice)
        self.redis.hset(self._redis_key(), "choices", escape.json_encode(choices))

    def remove_choice(self, choice):
        """Adds a choice for the experiment"""

        if not ALLOWED_NAMES.match(choice):
            raise ExperimentException(self.name, "Illegal choice name: %s" % choice)

        choices = self.choices()
        choices.remove(choice)
        self.redis.hset(self._redis_key(), "choices", escape.json_encode(choices))

    def plays(self, choice):
        """Gets the play count associated with an experiment choice"""
        return int(self.redis.hget(self._redis_key(), "%s:plays" % choice) or 0)

    def add_play(self, choice, count=1):
        """Increments the play count for a given experiment choice"""
        self.redis.hincrby(self._redis_key(), "%s:plays" % choice, count)

    def rewards(self, choice):
        """Gets the reward count associated with an experiment choice"""
        return int(self.redis.hget(self._redis_key(), "%s:rewards" % choice) or 0)

    def add_reward(self, choice, count=1):
        """Increments the reward count for a given experiment choice"""
        self.redis.hincrby(self._redis_key(), "%s:rewards" % choice, count)

    def get_default_choice(self):
        """Gets the default choice for this experiment"""
        return self.redis.hget(self._redis_key(), "default-choice")

    def set_default_choice(self, choice):
        """Sets the default choice for this experiment"""
        self.redis.hset(self._redis_key(), "default-choice", choice)

    def results(self):
        """
        Gets all of the data associated with this experiment, and stores it in
        a JSON-serializable data structure
        """

        if not self.exists():
            raise ExperimentException(self.name, "does not exist")

        # Build a list of all the choices
        choices = []

        for choice in self.choices():
            plays = self.plays(choice)
            rewards = self.rewards(choice)
            results = rewards / (plays or 1)
            choices.append(dict(name=choice, plays=plays, rewards=rewards, results=results))

        # Find the top two choices
        choices.sort(key=lambda c: c["results"], reverse=True)

        # Set the default choice for this experiment
        if len(choices) >= 1:
            self.set_default_choice(choices[0]["name"])

        # Get the chi-squared between the top two choices, if more than two choices exist
        if len(choices) >= 2:
            chi_squared = get_chi_squared(choices[0]["plays"], choices[0]["rewards"], choices[1]["plays"], choices[1]["rewards"])
        else:
            chi_squared = None
        
        # Return the results
        return {
            "name": self.name,
            "metadata": self.metadata(),
            "default": self.get_default_choice(),
            "chi_squared": chi_squared,
            "choices": choices
        }

def get_experiments(redis, active=True):
    """Gets the full list of experiments"""

    key = ACTIVE_EXPERIMENTS_REDIS_KEY if active else ARCHIVED_EXPERIMENTS_REDIS_KEY
    return [Experiment(redis, name) for name in redis.smembers(key)]



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

from boto.s3.connection import S3Connection

import os
import hashlib
import mimetypes
import oz.app

from .actions import *
from .middleware import *
from .options import *
from .tests import *

def static_url(redis, path):
    """Gets the static path for a file"""
    file_hash = get_cache_buster(redis, path)
    return "%s/%s?v=%s" % (oz.app.settings["static_host"], path, file_hash)

def get_cache_buster(redis, path):
    """Gets the cache buster value for a given file path"""
    return redis.hget("cache-buster:v1", path)

def set_cache_buster(redis, path, hash):
    """Sets the cache buster value for a given file path"""
    redis.hset("cache-buster:v1", path, hash)

def remove_cache_buster(redis, path):
    """Removes the cache buster for a given file"""
    redis.hdel("cache-buster:v1", path)

def get_bucket(s3_bucket=None):
    """Gets a bucket from specified settings"""
    settings = oz.app.settings
    s3_bucket = s3_bucket or settings["s3_bucket"]
    return S3Connection(settings["aws_access_key"], settings["aws_secret_key"]).get_bucket(s3_bucket)

def get_file(path):
    if oz.app.settings["s3_bucket"]:
        bucket = get_bucket(oz.app.settings["s3_bucket"])
        return S3File(bucket.new_key(path))
    else:
        return LocalFile(oz.app.settings["static_path"], path)

class CDNFile(object):
    def __hash__(self):
        return hash("%s:%s" % (self.__class__.__name__, self.path()))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.path() == other.path()
        else:
            return False

    def hash(self, override=""):
        """Creates a cache buster value for the file"""
        return hashlib.md5(override + self.contents()).hexdigest()[:6]

    def path(self):
        """Gets the path of the file"""
        raise NotImplementError()

    def upload(self, contents, replace=False):
        """
        Uploads the file to its path with the given `content`, adding the
        appropriate parent directories when needed. If the path already exists
        and `replace` is `False`, the file will not be uploaded.
        """
        raise NotImplementError()

    def contents(self):
        """Gets the contents of the file"""
        raise NotImplementError()

    def exists(self):
        """Returns whether the path exists"""
        raise NotImplementError()

    def remove(self):
        """Removes the given file"""
        raise NotImplementError()

class LocalFile(CDNFile):
    """
    Specifies a file stored locally and hosted through Tornado's static asset
    engine
    """

    def __init__(self, static_path, file_path):
        self.static_path = static_path
        self.file_path = file_path
        self.full_path = os.path.join(self.static_path, self.file_path)

    def path(self):
        return self.file_path

    def upload(self, contents, replace=False):
        if replace or not os.path.exists(self.full_path):
            try:
                os.makedirs(os.path.dirname(self.full_path))
            except:
                pass

            with open(self.full_path, "wb") as f:
                f.write(contents)

    def contents(self):
        with open(self.full_path, "r") as f:
            return f.read()

    def exists(self):
        return os.path.exists(self.full_path)

    def remove(self):
        os.remove(self.full_path)

class S3File(CDNFile):
    """Specifies a file stored on Amazon S3"""

    def __init__(self, key):
        self.key = key

    def path(self):
        return self.key.name

    def upload(self, contents, replace=False):
        if replace or not self.key.exists():
            guessed_type = mimetypes.guess_type(self.path())[0]

            self.key.set_contents_from_string(contents, {
                "Content-Type": guessed_type or "binary/octet-stream",
                "Cache-Control": "'max-age=155520000, public'",
                "Expires": "'Sat, 29 Apr 2017 13:31:45-0000 GMT'"
            })

    def contents(self):
        return self.key.get_contents_as_string()

    def exists(self):
        return self.key.exists()

    def remove(self):
        self.key.delete()



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

from sqlalchemy import create_engine
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
import tornado.options
import oz
import oz.app

from .actions import *
from .middleware import *
from .options import *

Base = declarative_base()
engine = None
Session = None

@oz.signal("initialized")
def initialize(pool_timeout=None):
    """Initializes a new SQLAlchemy environment"""

    global engine, Session

    kwargs = { "echo": oz.app.settings["debug_sql"] }
    if pool_timeout != None: kwargs["pool_timeout"] = pool_timeout

    engine = create_engine(oz.app.settings["db"], **kwargs)
    Session = sessionmaker(bind=engine)

def setup():
    """Initializes the tables if they don't exist already"""
    return Base.metadata.create_all(engine)

def session():
    """Gets a SQLAlchemy session"""
    global Session
    assert Session != None, "SQLAlchemy has not been initialized"
    return Session()



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz.plugins.redis_sessions
import oz
import unittest

@oz.test
class CDNCoreTest(unittest.TestCase):
    def test_password_hash(self):
        hash = oz.plugins.redis_sessions.password_hash("bar", password_salt="foo")
        self.assertEqual(hash, "sha256!c3ab8ff13720e8ad9047dd39466b3c8974e592c2fa383d4a3960714caef0c4f2")

    def test_random_hex(self):
        hexchars = set("0123456789abcdef")

        s1 = oz.plugins.redis_sessions.random_hex(10)
        self.assertEqual(len(s1), 10)

        for c in s1:
            self.assertIn(c, hexchars)

        s2 = oz.plugins.redis_sessions.random_hex(10)
        self.assertEqual(len(s2), 10)

        for c in s2:
            self.assertIn(c, hexchars)

        self.assertNotEqual(s1, s2)



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz

oz.options(
    redis_cache_connections = dict(type=bool, default=True, help="Whether to cache the redis connection between requests to prevent TCP slow start"),
    redis_host = dict(type=str, default="localhost", help="Redis host"),
    redis_port = dict(type=int, default=6379, help="Redis port"),
    redis_db = dict(type=int, default=0, help="Redis database number"),
    redis_password = dict(type=str, default=None, help="Password to the redis database")
)


from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import tornado.web
import tornado.testing
import oz
import oz.app
import collections
import operator

# Pseudo-cookie storage
cookie_counter = 0
cookie_jar = collections.defaultdict(lambda: {})

# Pseudo-redis storage
redis_collection = collections.defaultdict(lambda: {})

class OzTestCase(tornado.testing.AsyncHTTPTestCase):
    """
    Base class for web tests that use the oz infrastructure. Override
    get_handlers to specify the endpoints used in the test.
    """

    forced_settings = {}

    def setUp(self):
        super(OzTestCase, self).setUp()

        self.old_settings = oz.app.settings

        new_settings = dict(oz.app.settings)
        new_settings.update(self.forced_settings)
        oz.app.settings = new_settings

    def tearDown(self):
        super(OzTestCase, self).tearDown()
        oz.app.settings = self.old_settings

    def get_app(self):
        self.app = tornado.web.Application(self.get_handlers(), **self.get_app_kwargs())
        return self.app

    def get_handlers(self):
        raise NotImplementedError()

    def get_app_kwargs(self):
        return oz.app.settings

class FakeCookiesHandler(oz.RequestHandler):
    """
    A request handler that has a pseudo-cookie storage engine. Used because
    the HTTP client used for tornado unit testing doesn't persist cookies
    across requests.
    """

    def initialize(self, *args, **kwargs):
        super(FakeCookiesHandler, self).initialize(*args, **kwargs)
        self.__cookie_id = None

    @property
    def _cookie_id(self):
        """
        The cookie ID associated with this request. This value will be unique
        for the request.
        """

        if not self.__cookie_id:
            # Generate a cookie ID if one hasn't been made yet
            self.__cookie_id = self.get_argument("cookie_id", None)

            if not self.__cookie_id:
                global cookie_counter
                self.__cookie_id = cookie_counter
                cookie_counter += 1

        return self.__cookie_id

    @property
    def cookies(self):
        return cookie_jar[self._cookie_id]

    def get_cookie(self, name, default=None):
        return cookie_jar[self._cookie_id].get(name, default)

    def set_cookie(self, name, value, **kwargs):
        cookie_jar[self._cookie_id][name] = value

    def clear_cookie(self, name):
        del cookie_jar[self._cookie_id][name]

    def clear_all_cookies(self):
        cookie_jar[self._cookie_id] = {}



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import tornado.web

from .middleware import *
from .options import *
from .tests import *

class ApiError(tornado.web.HTTPError):
    """Exception for API-based errors"""

    def __init__(self, message, code=400):
        super(ApiError, self).__init__(code)
        self.message = message


from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.plugins.redis

class RedisMiddleware(object):
    def redis(self):
        """Gets or creates a connection to the redis database"""
        return oz.plugins.redis.create_connection()



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.testing
import oz.plugins.json_api
from oz.plugins.json_api import ApiMiddleware
import tornado.escape

class ApiMiddlewareTestHandler(oz.RequestHandler, ApiMiddleware):
    def initialize(self, *args, **kwargs):
        super(ApiMiddlewareTestHandler, self).initialize(*args, **kwargs)

@oz.test
class ApiMiddlewareTest(oz.testing.OzTestCase):
    forced_settings = {
        "allow_jsonp": True
    }

    def get_handlers(self):
        class UnknownErrorWritingApiHandler(ApiMiddlewareTestHandler):
            def get(self):
                raise Exception("Test 1")

        class ApiErrorWritingApiHandler(ApiMiddlewareTestHandler):
            def get(self):
                raise oz.plugins.json_api.ApiError("Test 2", code=401)

        class ResponseApiHandler(ApiMiddlewareTestHandler):
            def get(self):
                self.respond({"hello": "world"})

        class BodyApiHandler(ApiMiddlewareTestHandler):
            def put(self):
                self.respond(self.body())

        return [
            ("/unknown_error_writer", UnknownErrorWritingApiHandler),
            ("/api_error_writer", ApiErrorWritingApiHandler),
            ("/response", ResponseApiHandler),
            ("/body", BodyApiHandler),
        ]

    def test_error_writer(self):
        self.http_client.fetch(self.get_url('/unknown_error_writer'), self.stop)
        response = self.wait()
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertTrue("Exception: Test 1" in response.body)

        self.http_client.fetch(self.get_url('/api_error_writer'), self.stop)
        response = self.wait()
        json = tornado.escape.json_decode(response.body)
        self.assertEqual(response.code, 401)
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertEqual(json["code"], 401)
        self.assertEqual(json["error"], "Test 2")
        self.assertTrue("trace" in json)

    def test_respond(self):
        # Checks w/ JSONP support
        self.http_client.fetch(self.get_url('/response'), self.stop)
        response = self.wait()
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertEqual(response.body, '{"hello": "world"}')

        self.http_client.fetch(self.get_url('/response?callback=foo'), self.stop)
        response = self.wait()
        self.assertEqual(response.headers["Content-Type"], "application/javascript; charset=UTF-8")
        self.assertEqual(response.body, 'foo({"hello": "world"})')

        self.http_client.fetch(self.get_url('/response?callback=!!!'), self.stop)
        response = self.wait()
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertTrue("Invalid callback identifier" in response.body)

    def test_body(self):
        # Invalid: no content-type specified
        self.http_client.fetch(self.get_url('/body'), self.stop, body='{"hello": "world"}', method="PUT")
        response = self.wait()
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertTrue("JSON body expected" in response.body, msg="Unexpected body: %s" % response.body)

        self.http_client.fetch(self.get_url('/body'), self.stop, body='{"hello": "world"}', method="PUT", headers={"Content-Type": "application/json"})
        response = self.wait()
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertEqual(response.body, '{"hello": "world"}')

@oz.test
class NoJSONPApiMiddlewareTest(oz.testing.OzTestCase):
    forced_settings = {
        "allow_jsonp": False
    }

    def get_handlers(self):
        class NoJSONPResponseApiHandler(ApiMiddlewareTestHandler):
            def get(self):
                self.respond({"hello": "world"})

        return [
            ("/nojsonp_response", NoJSONPResponseApiHandler),
        ]

    def test_respond(self):
        # Checks w/o JSONP support
        self.http_client.fetch(self.get_url('/nojsonp_response'), self.stop)
        response = self.wait()
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertEqual(response.body, '{"hello": "world"}')

        self.http_client.fetch(self.get_url('/nojsonp_response?callback=foo'), self.stop)
        response = self.wait()
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertEqual(response.body, '{"hello": "world"}')

        self.http_client.fetch(self.get_url('/nojsonp_response?callback=!!!'), self.stop)
        response = self.wait()
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertEqual(response.body, '{"hello": "world"}')

@oz.test
class NoDebugApiMiddlewareTest(oz.testing.OzTestCase):
    forced_settings = {
        "debug": False
    }

    def get_handlers(self):
        class NoDebugErrorWritingApiHandler(ApiMiddlewareTestHandler):
            def get(self):
                raise oz.plugins.json_api.ApiError("Test 3", code=401)

        return [
            ("/nodebug_api_error_writer", NoDebugErrorWritingApiHandler),
        ]

    def test_error_writer(self):
        self.http_client.fetch(self.get_url('/nodebug_api_error_writer'), self.stop)
        response = self.wait()
        json = tornado.escape.json_decode(response.body)
        self.assertEqual(response.code, 401)
        self.assertEqual(response.headers["Content-Type"], "application/json; charset=UTF-8")
        self.assertEqual(json["code"], 401)
        self.assertEqual(json["error"], "Test 3")
        self.assertTrue("trace" not in json)



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz.app
import inspect
import oz

def explore_dict(d):
    if len(d):
        for value in d.values():
            doc = getattr(value, "__doc__")

            if doc:
                print("- %s: %s" % (value.__name__, doc))
            else:
                print("- %s" % value.__name__)
    else:
        print("  None")

def get_classes(module, subclass_of):
    for _, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and issubclass(obj, subclass_of):
            yield obj

def explore(plugin_name):
    """
    Explores a given plugin, printing out its actions, UIModules, middleware,
    routes, options and tests
    """

    module = __import__(plugin_name, globals(), locals(), [], -1)

    for sub_module_name in plugin_name.split(".")[1:]:
       module = getattr(module, sub_module_name)

    print("Actions")
    explore_dict(oz.app.actions)

    print("\nUIModules")
    explore_dict(oz.app.uimodules)

    print("\nRoutes")
    if oz.app.routes:
        for route in oz.app.routes:
            print("- %s" % route[0])
            print("  - request handler: %s" % route[1])

            if len(route) > 2:
                print("  - request handler initialization args: %s" % route[2])
    else:
        print("  None")

    print("\nOptions")
    if oz.app.options:
        for option_name, option_params in oz.app.options.items():
            print("- %s" % option_name)

            for option_param_name, option_param in option_params.items():
                print("  - %s: %s" % (option_param_name, option_param))
    else:
        print("  None")

    print("\nTests")
    if oz.app.tests:
        for test in oz.app.tests:
            print("- %s" % test.__name__)
    else:
        print("  None")



from setuptools import setup

version = "0.1"

setup(
    name = "Oz",
    version = version,
    description = "A batteries-included web framework built on Tornado",
    author = "Yusuf Simonson",
    author_email = 'yusuf@themuse.com',
    url = "http://github.com/dailymuse/oz",
    zip_safe = False,
    
    packages = [
        "oz",
        "oz.admin_actions",
        "oz.plugins",
        "oz.plugins.aws_cdn",
        "oz.plugins.aws_cdn.tests",
        "oz.plugins.bandit",
        "oz.plugins.bandit.tests",
        "oz.plugins.blinks",
        "oz.plugins.blinks.tests",
        "oz.plugins.core",
        "oz.plugins.json_api",
        "oz.plugins.json_api.tests",
        "oz.plugins.redis",
        "oz.plugins.redis_sessions",
        "oz.plugins.redis_sessions.tests",
        "oz.plugins.sqlalchemy",
    ],
    
    scripts = ["scripts/oz"],

    package_data = {
        "oz": [
            "plugins/bandit/assets/*",
            "skeleton/*.py",
            "skeleton/plugin/*.py",
            "skeleton/plugin/tests/*.py",
        ]
    },

    install_requires = [
        "tornado>=2.4.1",
        "optfn==0.4.1"
    ],

    extras_require = {
        "oz.plugins.aws_cdn": ["boto>=2.9.7"],
        "oz.plugins.redis": ["redis>=2.6.0"],
        "oz.plugins.sqlalchemy": ["sqlalchemy>=0.7.8"]
    }
)



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.app
import oz.plugins.redis_sessions

class RedisSessionMiddleware(object):
    def _session_key(self):
        """Gets the redis key for a session"""

        session_id = self.get_cookie("session_id")

        # Generate a new session if one does not exist yet
        if session_id == None:
            session_id = oz.plugins.redis_sessions.random_hex(20)
            self.set_cookie("session_id", session_id, httponly=True)

        password_salt = oz.app.settings["session_salt"]
        return "session:%s:v2" % oz.plugins.redis_sessions.password_hash(password_salt, session_id)

    def _update_session_expiration(self):
        """
        Updates a redis item to expire later since it has been interacted with
        recently
        """

        session_time = oz.app.settings["session_time"]

        if session_time:
            self.redis().expire(self._session_key(), session_time)

    def get_session_value(self, name, default=None):
        """Gets a session value"""

        value = self.redis().hget(self._session_key(), name) or default
        self._update_session_expiration()
        return value

    def set_session_value(self, name, value):
        """Sets a session value"""

        self.redis().hset(self._session_key(), name, value)
        self._update_session_expiration()

    def clear_session_value(self, name):
        """Removes a session value"""
        self.redis().hdel(self._session_key(), name)
        self._update_session_expiration()

    def clear_all_session_values(self):
        """Kills a session"""
        self.redis().delete(self._session_key())
        self.clear_cookie("session_id")



project_name = "__PROJECT_NAME__"

plugins = [
    "oz.plugins.core",
    project_name,
]

app_options = dict(
    port = 8000,
    debug = True,
    static_path = "static",
    template_path = "templates",
    xsrf_cookies = False,
    gzip = False
)



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.testing
import oz.plugins.bandit
from oz.plugins.bandit import BanditTestingMiddleware
from oz.plugins.redis_sessions import RedisSessionMiddleware
from oz.plugins.redis import RedisMiddleware

@oz.test
class BanditMiddlewareTest(oz.testing.OzTestCase):
    forced_settings = {
        "session_salt": "abc"
    }

    def get_handlers(self):
        class BanditHandler(oz.testing.FakeCookiesHandler, RedisMiddleware, RedisSessionMiddleware, BanditTestingMiddleware):
            def post(self):
                choice = self.choose_experiment("ex-middleware")
                self.finish(choice)

            def get(self):
                self.experiment_success("ex-middleware")
                choice = self.get_experiment_choice("ex-middleware")
                redis = self.redis()
                experiment = oz.plugins.bandit.Experiment(redis, "ex-middleware")
                self.finish("%s:%s" % (experiment.plays(choice), experiment.rewards(choice)))

        return [
            ("/bandit", BanditHandler),
        ]

    def tearDown(self):
        super(BanditMiddlewareTest, self).tearDown()
        
        # Clean up just in case there's keys still lying around
        redis = oz.plugins.redis.create_connection()

        for key in redis.keys("bandit:*:v2"):
            redis.delete(key)

    def test_bandit(self):
        redis = oz.plugins.redis.create_connection()

        experiment = oz.plugins.bandit.Experiment(redis, "ex-middleware")
        experiment.add()

        for choice in ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]:
            experiment.add_choice(choice)

        # Add an empty-ish body because tornado's simple_httpclient requires it
        self.http_client.fetch(self.get_url("/bandit?cookie_id=test_bandit"), self.stop, method="POST", body=" ")
        response = self.wait()
        choice = response.body
        self.assertIn(choice, ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"])

        self.http_client.fetch(self.get_url("/bandit?cookie_id=test_bandit"), self.stop, method="POST", body=" ")
        response = self.wait()
        self.assertEqual(response.body, choice)

        self.http_client.fetch(self.get_url("/bandit?cookie_id=test_bandit"), self.stop)
        response = self.wait()
        self.assertEqual(response.body, "2:1")



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz

oz.options(
    db = dict(type=str, help="SQLAlchemy connection string to the database"),
    debug_sql = dict(type=bool, default=False, help="Dump SQL queries performed by SQLAlchemy to stdout"),
)


from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import random
import oz.plugins.bandit

class BanditTestingMiddleware(object):
    def __init__(self):
        super(BanditTestingMiddleware, self).__init__()
        self.template_helper("get_experiment_choice", self.get_experiment_choice)
        self.template_helper("choose_experiment", self.choose_experiment)

    def session_key(self, experiment):
        """Gets the session key for an experiment"""
        return "bandit:%s:v1" % experiment

    def join_experiment(self, experiment, choice):
        """Opts the current user into an experiment choice"""
        self.set_session_value(self.session_key(experiment), choice)

    def leave_experiment(self, experiment):
        """Boots the current user out of an experiment"""
        self.clear_session_value(self.session_key(experiment))

    def get_experiment_choice(self, experiment):
        """Gets the experiment choice a user is in"""
        return self.get_session_value(self.session_key(experiment))

    def choose_experiment(self, name):
        """
        Opts a user into one of many choices for an experiment. They have a
        90 percent chance of being opted into the default choice (the one with
        the greatest results.) Otherwise they join a random choice (including
        potentially the default choice anyway.)
        """

        redis = self.redis()
        experiment = oz.plugins.bandit.Experiment(redis, name)
        experiment_choices = experiment.choices()
        choice = self.get_experiment_choice(name)

        # If the currently selected user choice is no longer valid, nullify it
        if not choice in experiment_choices:
            choice = None

        if not choice:
            if random.random() >= 0.1:
                # Join the best performing choice
                default_choice = experiment.get_default_choice()

                if default_choice:
                    choice = default_choice

            if not choice and len(experiment_choices) > 0:
                # Join a random choice
                choice = random.choice(experiment_choices)

            # Opt the user in
            if choice:
                self.join_experiment(name, choice)

        # Add to the play count for the selected choice
        if choice:
            experiment.add_play(choice)
        
        return choice

    def experiment_success(self, name):
        """
        Marks the user's current choice as successful - i.e. the user
        converted, or clicked-through, etc.
        """
        redis = self.redis()
        choice = self.get_experiment_choice(name)
        experiment = oz.plugins.bandit.Experiment(redis, name)

        if choice:
            experiment.add_reward(choice)

    def choose_experiment_callback(self, name, *args, **kwargs):
        """
        Helper method that takes in a dict of choice names to callbacks. A
        choice is selected for the user, and the associated callback is
        executed with the specified `args` and `kwargs`. If the callback
        returns `True`, the choice is marked successful.
        """

        choice = self.choose_experiment(name)
        result = choices[choice](*args, **kwargs)
        if result == True: self.experiment_success(name)
        return choice
